/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */
package de.bsvrz.vew.bmvew.bmvew;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeSet;
import java.util.Vector;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataState;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.Data.ReferenceValue;
import de.bsvrz.dav.daf.main.Data.TextArray;
import de.bsvrz.dav.daf.main.Data.TextValue;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ClientApplication;
import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.DynamicObject;
import de.bsvrz.dav.daf.main.config.DynamicObjectType;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageCauser;
import de.bsvrz.sys.funclib.operatingMessage.MessageGrade;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;
import de.bsvrz.sys.funclib.operatingMessage.MessageState;
import de.bsvrz.sys.funclib.operatingMessage.MessageType;
import de.bsvrz.sys.funclib.dambach.vewdynobj.VerwaltungDynObj;
import de.bsvrz.vew.bmvew.bmvew.meldung.BenutzerAktion;
import de.bsvrz.vew.bmvew.bmvew.meldung.Meldung;
import de.bsvrz.vew.bmvew.bmvew.regel.Aktion;
import de.bsvrz.vew.bmvew.bmvew.regel.AusfuehrenVerteilen;
import de.bsvrz.vew.bmvew.bmvew.regel.Filter;
import de.bsvrz.vew.bmvew.bmvew.regel.Regel;
import de.bsvrz.vew.bmvew.bmvew.regel.RegelBasis;
import de.bsvrz.vew.bmvew.bmvew.regel.Aktion.Aktionen;
import de.bsvrz.vew.bmvew.bmvew.sender.Sender;
import de.bsvrz.vew.bmvew.bmvew.sender.SenderParameter;

/**
 * Die SW-Einheit Betriebsmeldungsverwaltung ist fr die Verwaltung der automatisch und manuell erzeugten Meldungen
 * zustndig. Die Betriebsmeldungsverwaltung prft die automatisch erstellten Meldungen (im folgenden
 * Informationsmeldungen genannt) nach parametrierbaren Kriterien und fhrt entsprechende Aktionen durch. Neben dieser
 * Behandlung der Informationsmeldungen ist die SWE fr die Verwaltung von Meldungen zustndig, die interaktiven
 * Charakter haben. Bei diesem Meldungstyp werden zustzlich Ablufe fr die Lebenszeit der Meldung verwaltet.<br>
 * 
 * <br>
 * Folgende Startparameter sind mglich:<br>
 * <br>
 * DAF-Startparameter wie in SSB beschrieben und durch StandardApplication verarbeitet:<br>
 * <br>
 * Beispiele:<br>
 * -datenverteiler=rechnerDav:8083<br>
 * -debugLevelStdErrText=CONFIG<br>
 * -debugLevelFileText=CONFIG<br>
 * -benutzer=Tester<br>
 * -authentifizierung=C:\kernsoftware-3.4.4\passwd<br>
 * 
 * <br>
 * Spezielle Startparameter:<br>
 * <br>
 * <ul>
 * <li>-konfigurationsbereich= hier wird die Pid des Konfigurationsbereichs uebergeben</li>
 * <br>
 * <li>-konfigurationsobjekt= hier wird die Pid des Konfigurationsobektes uebergeben</li>
 * <br>
 * <li>-groesseinfoanteil= hier wird die Groesse des Feldes des Informationsanteils uebergeben</li>
 * <br>
 * <li>-loeschoffset= hier wird die Zeit angegeben bis ein Meldungs-Objekt gelscht wird, nach dem sein Zustand auf
 * erledigt gesetzt worden ist</li>
 * <br>
 * <li>-versandobjekt= hier wird das Konfigrations-Objekt des Versandmoduls angegeben</li>
 * <br>
 * <li>-infokanaleingeschraenkterbetrieb= hier wird der InformationsKanal bergeben, auf dem die Betriebsmeldung
 * publiziert werden, wenn die Applikation nur eingeschrnkt betriebsbereit ist</li>
 * <br>
 * 
 * 
 * <br>
 * Beispiele:<br>
 * -konfigurationsbereich=kb.default.aoe.dambach.test1<br>
 * -konfigurationsobjekt=kv.aoe.dambach.test1<br>
 * -groesseInfoAnteil=3<br>
 * -loeschoffset=60 Sekunden<br>
 * -infokanaleingeschraenkterbetrieb=betriebsMeldung.informationsKanal<br>
 * <br>
 * 
 * @version $Revision: 1.5 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
 * 
 * @author Dambach-Werke GmbH
 * @author Timo Pittner
 * 
 */
public class BetriebsMeldungApp implements StandardApplication, ClientReceiverInterface, ClientSenderInterface,
    MutableSetChangeListener
{
  /**
   * DebugLogger fr Debug-Ausgaben
   */
  private static Debug _debug;

  /**
   * Verbindung zum Datenverteiler.
   */
  private static ClientDavInterface _connection;

  /**
   * Der Konfigurationsbereich.
   */
  private static ConfigurationArea _konfigurationsBereich;

  /**
   * Das Konfigurationsobjekt
   */
  private static ConfigurationObject _konfigurationsObjekt;

  /**
   * Das Versandobjekt
   */
  private static ConfigurationObject _versandObjektKex;

  /**
   * String fuer das Konfigurationsobjekt
   */
  private String _strKonfigurationsObjekt;

  /**
   * String fuer das Konfigurationsbereich
   */
  private String _strKonfigurationsBereich;

  /**
   * String fuer das VersandObjekt
   */
  private String _strVersandObjektKex;

  /**
   * String fuer das Informations-Objekt beim eingeschaenkten Betrieb
   */
  private String _strInfoKanalEingeschraenkterBetrieb;

  /**
   * Liste fr die Informationskanaele
   */
  private static List<SystemObject> _informationsKanaele;

  /**
   * Das Datenmodell
   */
  private DataModel _datenmodell;

  /**
   * Die statische Liste der Ereignisse
   * 
   */
  private static Map<String, Meldung> meldungListTmp = new HashMap<String, Meldung>();

  private static Map<String, Meldung> meldungList = Collections.synchronizedMap(meldungListTmp);

  /**
   * Der Zaehler fr die Meldungen
   */
  private int _cntMeldung;

  /**
   * Flag, das die Initialisierungsphase anzeigt
   */
  private boolean _inInit = false;

  /**
   * Speicher fr Daten zum Parametrieren.
   */
  private static Vector<ResultData> _inputDaten = new Vector<ResultData>();

  /**
   * Speicher fr Daten zum Senden.
   */
  private static Vector<ResultData> _sendeDaten = new Vector<ResultData>();

  /**
   * Speicher fr Daten zum Parametrieren.
   */
  private static Vector<ResultData> _paramDaten = new Vector<ResultData>();

  /**
   * Zeit in Millisekunden, nach der das Objekt geloescht wird
   */
  private static long _loeschOffset;

  /**
   * Die Groesse des Speichers der Informationsanteile.
   */
  private static int _groesseInfoAnteil;

  /**
   * Flag, das anzeigt ob nur ein eingeschraenkter Betrieb mglich ist
   */
  private int _eingeschraenkterBetrieb = 0;

  /**
   * Die statische Liste der Meldungsgruppenzuordnung
   * 
   */
  private static Map<String, Set<String>> _meldungsGruppenZuordnung = new HashMap<String, Set<String>>();

  /**
   * Die von der Applikation verwaltete laufende Nummer der versendeten Faxe.
   */
  private static long _laufendeNummerEmailFax = 0L;

  /**
   * Startet die Applikation.
   * 
   * @param args
   *          bergebene Parameter
   */
  public static void main(String[] args)
  {
    StandardApplicationRunner.run(new BetriebsMeldungApp(), args);
  }

  /**
   * Hier knnen die bergebenen Parameter ausgewertet werden.
   * 
   * @param argumentList
   * @throws Exception
   */
  public void parseArguments(ArgumentList argumentList) throws Exception
  {
    _debug = Debug.getLogger();

    _strKonfigurationsObjekt = argumentList.fetchArgument("-betriebsMeldungsVerwaltung=:aoe:").asNonEmptyString();

    if (_strKonfigurationsObjekt.equals(":aoe:"))
    {
      _strKonfigurationsObjekt = null;
    }

    _strKonfigurationsBereich = argumentList.fetchArgument("-konfigurationsbereich=").asNonEmptyString();

    _strVersandObjektKex = argumentList.fetchArgument("-versandobjektkex=").asNonEmptyString();

    _loeschOffset = argumentList.fetchArgument("-loeschoffset=").asRelativeTime();

    _groesseInfoAnteil = argumentList.fetchArgument("-groesseinfoanteil=").intValue();

    _strInfoKanalEingeschraenkterBetrieb = argumentList.fetchArgument("-infokanaleingeschraenkterbetrieb=").asString();

    argumentList.fetchUnusedArguments();
  }

  /**
   * Nach Erhalt der Verbindung zum Datenverteiler meldet sich die Applikation als Senke auf die Betriebsmeldungen an.
   * 
   * @param connection
   *          Verbindung zum Datenverteiler
   * @throws Exception
   */
  public void initialize(ClientDavInterface connection) throws Exception
  {
    try
    {
      _debug = Debug.getLogger();
      _connection = connection;
      _datenmodell = _connection.getDataModel();

      DataModel configuration = _connection.getDataModel();

      _konfigurationsBereich = configuration.getConfigurationArea(_strKonfigurationsBereich);

      if (_strKonfigurationsObjekt != null)
      {
        _konfigurationsObjekt = (ConfigurationObject)configuration.getObject(_strKonfigurationsObjekt);
      }
      else
      {
        _konfigurationsObjekt = (ConfigurationObject)_connection.getLocalConfigurationAuthority();
        _strKonfigurationsObjekt = _konfigurationsObjekt.getPid();
      }

      if (_konfigurationsObjekt == null || !_konfigurationsObjekt.isOfType("typ.betriebsMeldungsVerwaltung"))
      {
        final String errorMessage = "Es gibt kein passendes BetriebsMeldungsVerwaltungs-Objekt (typ.betriebsMeldungsVerwaltung). Applikation kann nicht gestartet werden.";
        // _debug.error(errorMessage);
        _connection.disconnect(true, errorMessage);
        this.wait(1000);
        System.exit(0);

        // throw new ExceptionInInitializerError(errorMessage);
      }

      _versandObjektKex = (ConfigurationObject)configuration.getObject(_strVersandObjektKex);

      if (_versandObjektKex == null || !_versandObjektKex.isOfType("typ.versandModulExterneMeldungen"))
      {
        final String errorMessage = "Es gibt kein passendes Versand-Objekt (typ.versandModulExterneMeldungen). Applikation kann nicht gestartet werden.";
        // _debug.error(errorMessage);
        _connection.disconnect(true, errorMessage);
        this.wait(1000);
        System.exit(0);
        // throw new ExceptionInInitializerError(errorMessage);
      }

      _debug.config("Die Verwaltung nutzt folgendes BetriebsMeldungsVerwaltungsObjekt: " + _konfigurationsObjekt);

      AttributeGroup atg = configuration.getAttributeGroup("atg.betriebsMeldung");
      Aspect asp = configuration.getAspect("asp.information");
      DataDescription dataDescription = new DataDescription(atg, asp);

      SystemObjectType sot = _connection.getDataModel().getType("typ.informationsKanal");

      _informationsKanaele = sot.getElements();

      if (_informationsKanaele == null)
      {
        final String errorMessage = "Es gibt kein passendes Objekt vom Typ (betriebsMeldung.informationsKanal). Applikation kann nicht gestartet werden.";
        _debug.error(errorMessage);
        throw new ExceptionInInitializerError(errorMessage);
      }

      for (SystemObject s : _informationsKanaele)
      {
        _debug.config("Anmeldung als Quelle fr: " + s.getPid() + " -> dd: " + dataDescription);

      }
      _connection.subscribeSender(this, _informationsKanaele, dataDescription, SenderRole.source());

      /** ********************************************************************************* */

      Thread sendeArbeiter = new Thread(new SendeArbeiter(_sendeDaten));
      sendeArbeiter.start();

      Thread paramArbeiter = new Thread(new SendeParamArbeiter(_paramDaten));
      paramArbeiter.start();

      Thread inputArbeiter = new Thread(new EmpfangsArbeiter(_inputDaten));
      inputArbeiter.start();

      /** ********************************************************************************* */

      _connection.subscribeReceiver(this, _konfigurationsObjekt, dataDescription, ReceiveOptions.normal(), ReceiverRole
          .drain());
      _debug.config("Die Senke fr Betriebsmeldungen ist bereit.");

      List<SystemObject> menge = null;

      _eingeschraenkterBetrieb = 1;

      ConfigurationArea ca = _datenmodell.getConfigurationArea("kb.tmVewBetriebGlobal");
      short av = ca.getActiveVersion();

      if (av < 5)
        _eingeschraenkterBetrieb = -1;

      try
      {
        menge = leseMenge(_konfigurationsObjekt.getPid(), "Meldungen");
      }
      catch (IllegalArgumentException e)
      {
        // TODO Auto-generated catch block
        _eingeschraenkterBetrieb = -2;

        _debug
            .error("Fehler Menge Meldungen: Betriebsmeldungsverwaltung arbeitet eingeschgrnkt als Verteiler fr Betriebsmeldungen auf Informationskanal: "
                + _strInfoKanalEingeschraenkterBetrieb);

      }

      if (_eingeschraenkterBetrieb >= 0)
      {

        _inInit = true;

        atg = configuration.getAttributeGroup("atg.betriebsMeldungsVerwaltungRegel");
        asp = configuration.getAspect("asp.parameterSoll");
        dataDescription = new DataDescription(atg, asp);

        _connection.subscribeReceiver(this, _konfigurationsObjekt, dataDescription, ReceiveOptions.normal(),
            ReceiverRole.receiver());
        _debug.fine("Anmeldung auf Regel-Parameter fuer: " + _konfigurationsObjekt.getPid());

        atg = configuration.getAttributeGroup("atg.meldung");
        asp = configuration.getAspect("asp.parameterSoll");
        dataDescription = new DataDescription(atg, asp);

        for (SystemObject so : menge)
        {
          _debug.fine("Anmeldung auf Parameter fuer Meldung: " + so.getPid());

        }

        _cntMeldung = 0;

        _connection.subscribeReceiver(this, menge, dataDescription, ReceiveOptions.normal(), ReceiverRole.receiver());
 
        final List<SystemObject> fmenge = menge;
        TimerTask task = new TimerTask() {

          @Override
          public void run()
          {            
            if (_cntMeldung < fmenge.size())
            {
              final String errorMessage = "Es konnten keine/nicht alle Meldungsparameter eingelesen werden.";
              _debug.error(errorMessage);
              _connection.disconnect(true, errorMessage);
              System.exit(0);
            }
            else
            {
              _debug.fine("Kein Fehler nach Wartezeit: " + _cntMeldung);
            }
          }
        };

        Timer timer = new Timer();
        timer.schedule(task, 60000);

        synchronized (this)
        {
          while (_cntMeldung < menge.size())
          {
            this.wait();
          }
        }
       
        
        for (Map.Entry<String, Meldung> me : meldungList.entrySet())
        {
          SystemObject so = _connection.getDataModel().getObject(me.getKey());
          Meldung meldung = me.getValue();
          Data data = meldung.getData();

          atg = configuration.getAttributeGroup("atg.meldung");
          asp = configuration.getAspect("asp.zustand");
          DataDescription dd = new DataDescription(atg, asp);

          // publiziere
          _debug.fine("publiziere Daten fr Meldung: " + so.getPid());

          ResultData res = new ResultData(so, dd, System.currentTimeMillis(), data);

          synchronized (_sendeDaten)
          {
            _sendeDaten.add(res);
            _sendeDaten.notify();
          }

          // Sender sender = new Sender(_connection, SenderRole.source());
          // sender.sende(so, atg, asp, data);

        }

        atg = configuration.getAttributeGroup("atg.meldungsBearbeitung");
        asp = configuration.getAspect("asp.vorgabe");
        dataDescription = new DataDescription(atg, asp);

        _connection.subscribeReceiver(this, _konfigurationsObjekt, dataDescription, ReceiveOptions.normal(),
            ReceiverRole.drain());

        _debug.config("Anmeldung als Senke fr Benutzeraktionen: " + atg.getPid());

        SystemObjectType type = _datenmodell.getType("typ.meldungsGruppe");

        menge = type.getElements();

        for (SystemObject so : menge)
        {
          _debug.fine("Anmeldung auf Parameter der MeldungsGruppen: " + so.getPid());
        }

        atg = configuration.getAttributeGroup("atg.meldungsGruppeMeldung");
        asp = configuration.getAspect("asp.parameterSoll");
        dataDescription = new DataDescription(atg, asp);

        _connection.subscribeReceiver(this, menge, dataDescription, ReceiveOptions.normal(), ReceiverRole.receiver());

        atg = configuration.getAttributeGroup("atg.meldungsGruppeZuordnung");
        asp = configuration.getAspect("asp.vorgabe");
        dataDescription = new DataDescription(atg, asp);

        _connection.subscribeReceiver(this, _konfigurationsObjekt, dataDescription, ReceiveOptions.normal(),
            ReceiverRole.drain());

        _debug.config("Anmeldung als Senke fr Meldungsgruppen: " + atg.getPid());

        MutableSet ms = _konfigurationsObjekt.getMutableSet("Meldungen");

        if (ms != null)
        {
          _debug.config("Listener Menge Meldungen angemeldet");
          ms.addChangeListener(this);
        }

        atg = configuration.getAttributeGroup("atg.versandStatus");
        asp = configuration.getAspect("asp.auftragsStatus");
        dataDescription = new DataDescription(atg, asp);

        _connection.subscribeReceiver(this, _versandObjektKex, dataDescription, ReceiveOptions.normal(), ReceiverRole
            .receiver());

        _debug.config("Anmeldung als Empfnger fr den VersandStatus : " + atg.getPid());

        // Thread sendeArbeiter = new Thread(new SendeArbeiter(_sendeDaten));
        // sendeArbeiter.start();
        //        
        // Thread paramArbeiter = new Thread(new SendeParamArbeiter(_paramDaten));
        // paramArbeiter.start();

        _inInit = false;

        _debug.config("Betriebsmeldungsverwaltung bereit");

      }
      else
      {

        String s = "";
        if (_eingeschraenkterBetrieb == -1)
          s = "Fehler -> Version Konfigurationsbereich zu alt";
        else if (_eingeschraenkterBetrieb == -2)
          s = "Fehler -> konnte Menge Meldungen nicht lesen";
        else
          s = "nicht definiert";

        _debug.error("Betriebsmeldungsverwaltung eingeschraenkt bereit: " + s);

        ClientApplication app = _connection.getLocalApplicationObject();

        SystemObject user = _connection.getDataModel().getObject("");
        MessageCauser causer = new MessageCauser(user, "Fehler Betriebsmeldungsverwaltung", app.getName());
        MessageSender instance = MessageSender.getInstance();

        instance.setApplicationLabel(app.getName());

        Long id = app.getId();

        instance.sendMessage(id.toString(), MessageType.SYSTEM_DOMAIN, "", MessageGrade.ERROR,
            (SystemObject)getKonfigurationsObjekt(), MessageState.NEW_MESSAGE, causer, s);

      }

    }
    catch (ExceptionInInitializerError e)
    {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

  }

  /*
   * (non-Javadoc)
   * 
   * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
   */
  public void update(ResultData[] results)
  {
    // TODO Auto-generated method stub
    for (int i = 0; i < results.length; i++)
    {
      ResultData result = results[i];

      if (result != null)
      {
        synchronized (_inputDaten)
        {
          _inputDaten.add(result);
          _inputDaten.notify();
        }
      }
      else
      {
        // _debug.fine("result ist null");
      }
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#dataRequest(de.bsvrz.dav.daf.main.config.SystemObject,
   * de.bsvrz.dav.daf.main.DataDescription, byte)
   */
  public void dataRequest(SystemObject object, DataDescription dataDescription, byte state)
  {
    // TODO Auto-generated method stub
  }

  /*
   * (non-Javadoc)
   * 
   * @see de.bsvrz.dav.daf.main.ClientSenderInterface#isRequestSupported(de.bsvrz.dav.daf.main.config.SystemObject,
   * de.bsvrz.dav.daf.main.DataDescription)
   */
  public boolean isRequestSupported(SystemObject object, DataDescription dataDescription)
  {
    // TODO Auto-generated method stub
    return false;
  }

  /**
   * Liest alle Objekte aus der Menge der bergegbenen Pid
   * 
   * @param pidKonfigObjekt
   *          String Konfigurationsobjet
   * @param pidMenge
   *          String Menge
   * @return Liste der in Menge entghaltenen Objekte
   * 
   * @throws Exception
   */
  public List<SystemObject> leseMenge(String pidKonfigObjekt, String pidMenge) throws IllegalArgumentException
  {

    ConfigurationObject kalender = (ConfigurationObject)_connection.getDataModel().getObject(pidKonfigObjekt);

    ObjectSet objekte = kalender.getObjectSet(pidMenge);

    if (objekte == null)
    {

      throw new IllegalArgumentException();

    }

    List<SystemObject> list = objekte.getElements();
    
    List<SystemObject> tmp = new ArrayList<SystemObject>();

    tmp.addAll(list);

    for (SystemObject systemObject : list)
    {
      if (!systemObject.isValid())
      {
        _debug.fine("ungueltiges Objekt wird aus Menge entfernt: " + systemObject);
        tmp.remove(systemObject);

        DynamicObjectType dot = (DynamicObjectType)_datenmodell.getType("typ.meldung");

        AttributeGroup attributgruppe = _datenmodell.getAttributeGroup("atg.meldung");
        Aspect aspekt = _datenmodell.getAspect("asp.eigenschaften");

        VerwaltungDynObj vew = new VerwaltungDynObj(_connection, _datenmodell, _konfigurationsBereich, dot,
            _konfigurationsObjekt, attributgruppe, aspekt);

        vew.setDynamicObject((DynamicObject)systemObject);
        vew.entferneVonMenge("Meldungen");

      }
    }

    return tmp;
  }

  public static Debug getDebug()
  {
    return _debug;
  }

  public static void setDebug(Debug debug)
  {
    _debug = debug;
  }

  public static List<SystemObject> getInformationsKanaele()
  {
    return _informationsKanaele;
  }

  public void setInformationsKanaele(List<SystemObject> informationsKanaele)
  {
    _informationsKanaele = informationsKanaele;
  }

  public static ClientDavInterface getConnection()
  {
    return _connection;
  }

  public static void setConnection(ClientDavInterface connection)
  {
    _connection = connection;
  }

  public static ConfigurationObject getKonfigurationsObjekt()
  {
    return _konfigurationsObjekt;
  }

  public static void setKonfigurationsObjekt(ConfigurationObject konfigurationsObjekt)
  {
    _konfigurationsObjekt = konfigurationsObjekt;
  }

  public static ConfigurationArea getKonfigurationsBereich()
  {
    return _konfigurationsBereich;
  }

  public static void setKonfigurationsBereich(ConfigurationArea konfigurationsBereich)
  {
    _konfigurationsBereich = konfigurationsBereich;
  }

  public static Map<String, Meldung> getMeldungList()
  {
    return meldungList;
  }

  public static void setMeldungList(Map<String, Meldung> meldungList)
  {
    BetriebsMeldungApp.meldungList = meldungList;
  }

  /*
   * (non-Javadoc)
   * 
   * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet,
   * de.bsvrz.dav.daf.main.config.SystemObject[], de.bsvrz.dav.daf.main.config.SystemObject[])
   */
  public void update(MutableSet set, SystemObject[] addedObjects, SystemObject[] removedObjects)
  {
    // TODO Auto-generated method stub
    _debug.fine("update Menge: " + set.getName());

    if (set.getName().equals("Meldungen"))
    {

      if (addedObjects != null)
      {
        List<SystemObject> list = new ArrayList<SystemObject>();

        for (int i = 0; i < addedObjects.length; i++)
        {

          SystemObject so = addedObjects[i];

          if (!meldungList.containsKey(so.getPid()))
          {
            _debug.fine(so.getPid() + " : Objekt nicht in Liste vorhanden!");
          }
          else
          {
            _debug.fine(so.getPid() + " : Objekt schon in Liste vorhanden!");
          }

          _debug.fine("added - " + addedObjects[i]);

          list.add(so);

        }

        _debug.fine("Anmelden auf " + list.size() + " Daten!!!");

        AttributeGroup atg = _connection.getDataModel().getAttributeGroup("atg.meldung");
        Aspect asp = _connection.getDataModel().getAspect("asp.parameterSoll");
        DataDescription dd = new DataDescription(atg, asp);
        _connection.subscribeReceiver(this, list, dd, ReceiveOptions.normal(), ReceiverRole.receiver());

      }

      if (removedObjects != null)
      {
        for (int i = 0; i < removedObjects.length; i++)
        {
          SystemObject so = removedObjects[i];

          if (meldungList.containsKey(so.getPid()))
          {
            meldungList.remove(so.getPid());
            _debug.fine(so.getPid() + " : Objekt geloescht!");
          }
          else
          {
            _debug.fine(so.getPid() + " : Objekt nicht verhanden!");
          }
          _debug.fine("removed - " + removedObjects[i]);

        }

      }

    }

  }

  /**
   * Innere Klasse fr die Abarbeitung der zu sendenden Daten. Die Abarbeitung wird als Thread gestartet.
   * 
   * @version $Revision: 1.5 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
   * 
   * @author Dambach-Werke GmbH
   * @author Timo Pittner
   * 
   */
  class SendeArbeiter implements Runnable
  {
    Vector<ResultData> results;

    public SendeArbeiter(Vector<ResultData> results)
    {
      this.results = results;
    }

    public void run()
    {
      while (true)
      {
        if (results.size() > 0)
        {
          datenVersenden();
        }
        else
        {
          try
          {
            synchronized (results)
            {
              results.wait();
            }
          }
          catch (InterruptedException e)
          {
            e.printStackTrace();
          }
        }
      }
    }
  }

  /**
   * Methode zur Verarbeitung der zu sendenden Daten.
   */
  private void datenVersenden()
  {

    Sender sender = new Sender(_connection);

    while (true)
    {
      ResultData resultData = null;

      synchronized (_sendeDaten)
      {
        if (_sendeDaten.size() > 0)
        {
          resultData = (ResultData)_sendeDaten.remove(0);

          SystemObject so = resultData.getObject();
          AttributeGroup atg = resultData.getDataDescription().getAttributeGroup();
          Aspect asp = resultData.getDataDescription().getAspect();
          Data data = resultData.getData();

          if (asp.equals(_connection.getDataModel().getAspect("asp.zustand")))
            sender.sende(so, atg, asp, data, SenderRole.source());
          else
            sender.sende(so, atg, asp, data, SenderRole.sender());

        }
      }

      if (resultData == null)
        break;
    }
  }

  public static Vector<ResultData> getSendeDaten()

  {
    return _sendeDaten;
  }

  /**
   * Innere Klasse fr die Abarbeitung der zu sendenden Parameterdaten. Die Abarbeitung wird als Thread gestartet.
   * 
   * @version $Revision: 1.5 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
   * 
   * @author Dambach-Werke GmbH
   * @author Timo Pittner
   * 
   */
  class SendeParamArbeiter implements Runnable
  {
    Vector<ResultData> params;

    public SendeParamArbeiter(Vector<ResultData> params)
    {
      this.params = params;
    }

    public void run()
    {
      while (true)
      {
        if (params.size() > 0)
        {
          paramDatenVersenden();
        }
        else
        {
          try
          {
            synchronized (params)
            {
              params.wait();
            }
          }
          catch (InterruptedException e)
          {
            e.printStackTrace();
          }
        }
      }
    }
  }

  /**
   * Methode zur Verarbeitung der zu sendenden Parameterdaten.
   */
  private void paramDatenVersenden()
  {

    SenderParameter senderParam = new SenderParameter(_connection, SenderRole.sender());

    while (true)
    {
      ResultData resultData = null;

      synchronized (_paramDaten)
      {
        if (_paramDaten.size() > 0)
        {
          resultData = (ResultData)_paramDaten.remove(0);

          SystemObject so = resultData.getObject();
          AttributeGroup atg = resultData.getDataDescription().getAttributeGroup();
          Aspect asp = resultData.getDataDescription().getAspect();
          Data data = resultData.getData();

          senderParam.sende(so, atg, asp, data);
        }
      }

      if (resultData == null)
        break;
    }
  }

  public static Vector<ResultData> getParamDaten()
  {
    return _paramDaten;
  }

  /**
   * Innere Klasse fr die Abarbeitung der empfangenen Datenstze. Die Abarbeitung wird als Thread gestartet.
   * 
   * @version $Revision: 1.5 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
   * 
   * @author Dambach-Werke GmbH
   * @author Timo Pittner
   * 
   */
  class EmpfangsArbeiter implements Runnable
  {
    Vector<ResultData> inputs;

    public EmpfangsArbeiter(Vector<ResultData> inputs)
    {
      this.inputs = inputs;
    }

    public void run()
    {
      while (true)
      {
        if (inputs.size() > 0)
        {
          empfangeneDatenVerarbeiten();
        }
        else
        {
          try
          {
            synchronized (inputs)
            {
              inputs.wait();
            }
          }
          catch (InterruptedException e)
          {
            e.printStackTrace();
          }
        }
      }
    }
  }

  /**
   * Methode zur Verarbeitung der empfangenen Daten.
   */
  private void empfangeneDatenVerarbeiten()
  {
    while (true)
    {
      ResultData result = null;

      if (_inputDaten.size() > 0)
      {
        result = (ResultData)_inputDaten.remove(0);

        String pid = result.getObject().getPid();

        final Data data = result.getData();

        if (data != null)
        {

          AttributeGroup atg1 = result.getDataDescription().getAttributeGroup();

          // Alle mglichen Attributgruppen im Kontext Betriebmeldungsverwaltung
          AttributeGroup atg2 = _connection.getDataModel().getAttributeGroup("atg.betriebsMeldung");
          AttributeGroup atg3 = _connection.getDataModel().getAttributeGroup("atg.betriebsMeldungsVerwaltungRegel");
          AttributeGroup atg4 = _connection.getDataModel().getAttributeGroup("atg.meldung");
          AttributeGroup atg5 = _connection.getDataModel().getAttributeGroup("atg.meldungsGruppeMeldung");
          AttributeGroup atg6 = _connection.getDataModel().getAttributeGroup("atg.meldungsBearbeitung");
          AttributeGroup atg7 = _connection.getDataModel().getAttributeGroup("atg.meldungsGruppeZuordnung");
          AttributeGroup atg8 = _connection.getDataModel().getAttributeGroup("atg.versandStatus");

          if (atg1.equals(atg2))
          {

            _debug.fine("Betriebsmeldung empfangen: " + data);

            if (_eingeschraenkterBetrieb >= 0)
            {

              RegelBasis.interpretiere(result);

              _debug.fine("Interpretiere Regel: " + result);

            }
            else
            {

              Data d = _connection.createData(atg3);
              Array a1 = d.getItem("Regel").asArray();
              a1.setLength(1);

              Array a2 = a1.getItem(0).getItem("Aktion").asArray();
              a2.setLength(1);
              TextValue tv = a2.getItem(0).getItem("Aktion").asTextValue();
              tv.setText(Aktionen.Verteilen.toString());

              TextArray ta = a2.getItem(0).getItem("Parameter").asTextArray();
              ta.setLength(1);

              if (_strInfoKanalEingeschraenkterBetrieb != null)
              {

                SystemObject so = _connection.getDataModel().getObject(_strInfoKanalEingeschraenkterBetrieb);

                if (so != null)
                {

                  if (so.isOfType(_connection.getDataModel().getType("typ.informationsKanal")))
                    ta.getTextValue(0).setText(_strInfoKanalEingeschraenkterBetrieb);
                  else
                    ta.getTextValue(0).setText("betriebsMeldung.informationsKanal");

                }
                else
                  ta.getTextValue(0).setText("betriebsMeldung.informationsKanal");

              }
              else
                ta.getTextValue(0).setText("betriebsMeldung.informationsKanal");

              Aktion aktion = new Aktion();
              aktion.initialisiere(tv, ta);
              AusfuehrenVerteilen verteilen = new AusfuehrenVerteilen(ta, tv);

              _debug.config("Ausfhren eingeschrnkt: " + result);

              verteilen.ausfuehrenAktion(result);
            }

          }
          else if (atg1.equals(atg3))
          {

            _debug.fine("Parameter atg.betriebsMeldungsVerwaltungRegel empfangen: " + data);

            Collection<Regel> tmpRegeln = new ArrayList<Regel>();
            Array arrayRegel = data.getItem("Regel").asArray();

            for (int loc1 = 0; loc1 < arrayRegel.getLength(); loc1++)
            {
              Array arrayFilter = arrayRegel.getItem(loc1).getItem("Filter").asArray();

              Filter[] filterFeld = new Filter[arrayFilter.getLength()];

              for (int loc2 = 0; loc2 < arrayFilter.getLength(); loc2++)
              {
                TextValue kriterium = arrayFilter.getItem(loc2).getItem("Kriterium").asTextValue();
                TextValue operation = arrayFilter.getItem(loc2).getItem("Operation").asTextValue();
                TextArray wert = arrayFilter.getItem(loc2).getItem("Wert").asTextArray();

                Filter filter = null;
                try
                {
                  filter = new Filter();
                  filter.initialisiere(kriterium, operation, wert);
                }
                catch (IllegalArgumentException e)
                {
                  // TODO Auto-generated catch block
                  // e.printStackTrace();
                  _debug.error("Fehler Filter, konnte Filter: " + loc2 + " nicht Initialisieren!");
                }

                filterFeld[loc2] = filter;

              }

              Array arrayAktion = arrayRegel.getItem(loc1).getItem("Aktion").asArray();

              Aktion[] aktionFeld = new Aktion[arrayAktion.getLength()];

              for (int loc3 = 0; loc3 < arrayAktion.getLength(); loc3++)
              {
                TextValue aktionen = arrayAktion.getItem(loc3).getItem("Aktion").asTextValue();
                TextArray parameter = arrayAktion.getItem(loc3).getItem("Parameter").asTextArray();

                Aktion aktion = null;
                try
                {
                  aktion = new Aktion();
                  aktion.initialisiere(aktionen, parameter);
                }
                catch (IllegalArgumentException e)
                {
                  // TODO Auto-generated catch block
                  // e.printStackTrace();
                  _debug.error("Fehler Aktion, konnte Aktion: " + loc3 + " nicht Initialisieren!");
                }

                aktionFeld[loc3] = aktion;

              }

              try
              {

                Regel regel = new Regel();
                // regel.initialisiere(_filter, _aktion);
                regel.initialisiere(filterFeld, aktionFeld);

                tmpRegeln.add(regel);
              }
              catch (IllegalArgumentException e)
              {
                // TODO Auto-generated catch block
                // e.printStackTrace();
                _debug.error("Fehler Regel, konnte Regel: " + loc1 + " nicht zur Regelbasis hinzufuegen!");
              }

            }

            RegelBasis.loescheRegelBasisListe();
            RegelBasis.getRegelBasisListe().addAll(tmpRegeln);

            _debug.fine("size Regeln: " + RegelBasis.getRegelBasisListe().size());

            for (Regel regel : RegelBasis.getRegelBasisListe())
            {
              Filter[] filters = regel.getFilters();
              for (int i = 0; i < filters.length; i++)
              {
                Filter filter = filters[i];
                _debug.fine(filter.toString());

              }

              Aktion[] aktionen = regel.getAktionen();
              for (int j = 0; j < aktionen.length; j++)
              {
                Aktion aktion = aktionen[j];
                _debug.fine(aktion.toString());
              }
            }
          }
          else if (atg1.equals(atg4))
          {
            _debug.fine("Parameter atg.meldung empfangen: " + data);

            if (!meldungList.containsKey(pid))
            {
              String text = result.getData().getItem("Wichtigkeit").asTextValue().getText();
              // if (!text.equals("keine nderung"))
              {
                // System.out.println("erzeuge : " + result);

                if (result.getData().getItem("InformationsMeldungsAnteil").asArray().getLength() > 0)
                {

                  Meldung meldung = new Meldung(result);

                  synchronized (meldungList)
                  {
                    meldungList.put(pid, meldung);
                    meldungList.notify();
                  }
                }
              }
            }
            else
            {
              _debug.fine("Parameter empfangen->Meldung: " + pid + " schon vorhanden");
            }

            if (_inInit)
            {

              _cntMeldung++;

              synchronized (this)
              {                 
                this.notify();
              }

            }

          }
          else if (atg1.equals(atg5))
          {

            _debug.fine("Parameter atg.meldungsGruppeMeldung empfangen: " + data);

            SystemObject so = result.getObject();

            Array a = data.getItem("MeldungReferenz").asArray();

            Set<String> set = new TreeSet<String>();

            for (int i = 0; i < a.getLength(); i++)
            {
              ReferenceValue ref = a.getItem(i).asReferenceValue();

              Meldung meldung = meldungList.get(ref.getSystemObjectPid());

              if (meldung != null)
              {
                meldung.getMeldungsGruppen().add(so.getPid());
                set.add(ref.getSystemObjectPid());
              }

            }

            _meldungsGruppenZuordnung.put(so.getPid(), set);
          }
          else if (atg1.equals(atg6))
          {

            _debug.fine("Benutzeraktion atg.meldungsBearbeitung empfangen: " + data);

            try
            {
              // TODO Auto-generated method stub
              BenutzerAktion ba = new BenutzerAktion();
              ba.auswerteBenutzerAktion(result);
            }
            catch (Exception e)
            {
              // TODO: handle exception
              _debug.error("Fehler beim Auswerten einer Benutzeraktion");
            }

          }
          else if (atg1.equals(atg7))
          {

            _debug.fine("Meldungsgruppenzuordnung atg.meldungsGruppeZuordnung empfangen: " + data);

            Array meldungen = data.getItem("MeldungReferenz").asArray();

            Array gruppen = data.getItem("MeldungsGruppeReferenz").asArray();

            Map<String, Set<String>> meldungsGruppenZuordnungNeu = new HashMap<String, Set<String>>();

            String pidGruppeTmp = null;
            String pidMeldungTmp = null;

            for (int j = 0; j < gruppen.getLength(); j++)
            {
              pidGruppeTmp = gruppen.getItem(j).asReferenceValue().getSystemObjectPid();

              Set<String> list = new HashSet<String>();

              for (int i = 0; i < meldungen.getLength(); i++)
              {
                pidMeldungTmp = meldungen.getItem(i).asReferenceValue().getSystemObjectPid();
                list.add(pidMeldungTmp);
                Meldung meldung = meldungList.get(pidMeldungTmp);
                meldung.getMeldungsGruppen().add(pidMeldungTmp);

              }

              if (pidGruppeTmp != null)
                meldungsGruppenZuordnungNeu.put(pidGruppeTmp, list);

            }

            Map<String, Set<String>> meldungsGruppenAenderung = new HashMap<String, Set<String>>();

            for (Map.Entry<String, Set<String>> me : meldungsGruppenZuordnungNeu.entrySet())
            {

              if (!_meldungsGruppenZuordnung.containsKey(me.getKey()))
              {
                meldungsGruppenAenderung.put(me.getKey(), me.getValue());
              }
              else
              {

                Set<String> set1 = _meldungsGruppenZuordnung.get(me.getKey());

                Set<String> set2 = me.getValue();

                boolean aenderung = false;

                if (set1.addAll(set2))
                  aenderung = true;

                // if (aenderung)
                {
                  meldungsGruppenAenderung.put(me.getKey(), set1);
                }
              }

            }

            // nderung in der Verwaltung der Meldungsgruppen:
            // Es werden nur Meldungen aus der Meldungsgruppe entfernt, wenn es keine Zuordnung
            // zu der Meldungsgruppe durch die Attributgruppe MeldungsGruppeZuordnung gibt.

            Set<String> vorher = _meldungsGruppenZuordnung.keySet();
            Set<String> neu = meldungsGruppenAenderung.keySet();
            Set<String> set = new HashSet<String>();

            for (int i = 0; i < meldungen.getLength(); i++)
            {
              pid = meldungen.getItem(i).asReferenceValue().getSystemObjectPid();
              set.add(pid);
            }
            for (String s : vorher)
            {
              if (!neu.contains(s))
              {
                Set<String> set1 = _meldungsGruppenZuordnung.get(s);

                boolean aenderung = false;
                HashSet<String> setNeu = new HashSet<String>();

                for (String s2 : set1)
                {
                  if (!set.contains(s2))
                    setNeu.add(s2);
                  else
                  {
                    aenderung = true;
                  }
                }
                if (aenderung)
                  meldungsGruppenAenderung.put(s, setNeu);
              }
            }

            for (Map.Entry<String, Set<String>> me : meldungsGruppenAenderung.entrySet())
            {
              AttributeGroup atg = _datenmodell.getAttributeGroup("atg.meldungsGruppeMeldung");
              Aspect asp = _datenmodell.getAspect("asp.zustand");
              DataDescription dd = new DataDescription(atg, asp);
              Data d = _connection.createData(atg);

              String key = me.getKey();
              Set<String> list = me.getValue();

              Array a = d.getItem("MeldungReferenz").asArray();

              a.setLength(list.size());

              int i = 0;
              for (String s : list)
              {
                SystemObject so = _connection.getDataModel().getObject(s);
                a.getItem(i).asReferenceValue().setSystemObject(so);
                i++;
              }

              SystemObject so = _connection.getDataModel().getObject(key);

              ResultData res1 = new ResultData(so, dd, System.currentTimeMillis(), d);

              synchronized (_sendeDaten)
              {
                _sendeDaten.add(res1);
                _sendeDaten.notify();
              }

              if (!_inInit)
              {

                asp = _datenmodell.getAspect("asp.parameterVorgabe");
                dd = new DataDescription(atg, asp);

                ResultData res2 = new ResultData(so, dd, System.currentTimeMillis(), d);

                synchronized (_paramDaten)
                {
                  _paramDaten.add(res2);
                  _paramDaten.notify();
                }

              }
              _meldungsGruppenZuordnung.remove(key);
              _meldungsGruppenZuordnung.put(key, list);
            }
          }
          else if (atg1.equals(atg8))
          {

            _debug.fine("VersandStatus atg.versandStatus empfangen: " + data);

            long idLocalAppObject = _connection.getLocalApplicationObject().getId();

            long idApplication = data.getItem("MeldungsID").getItem("ApplikationsID").asUnscaledValue().longValue();

            if (idLocalAppObject == idApplication)
            {

              String ok = data.getItem("VersandOk").asTextValue().getText();

              if (ok.equalsIgnoreCase("Ja"))
              {
                _laufendeNummerEmailFax += 1;
              }
              else
              {
                ClientApplication app = _connection.getLocalApplicationObject();
                SystemObject user = _connection.getDataModel().getObject("");
                MessageCauser causer = new MessageCauser(user, "Fehler Email-/Fax Versand", app.getName());
                MessageSender instance = MessageSender.getInstance();

                instance.setApplicationLabel(app.getName());

                Long id = app.getId();

                long l = data.getItem("MeldungsID").getItem("LaufendeNummer").asUnscaledValue().longValue();

                instance.sendMessage(id.toString(), MessageType.SYSTEM_DOMAIN, "", MessageGrade.ERROR,
                    (SystemObject)getKonfigurationsObjekt(), MessageState.NEW_MESSAGE, causer,
                    "Fax/Email konnte nicht versendet werden -> Lfd. Nummer: " + l);
              }
            }
          }
          else
          {
            _debug.error("Attributgruppe ist nicht definiert: " + data);
          }

        }

      }

      if (result == null)
        break;
    }
  }

  public static void setParamDaten(Vector<ResultData> paramDaten)
  {
    _paramDaten = paramDaten;
  }

  public static long getLoeschOffset()
  {
    return _loeschOffset;
  }

  public void setLoeschOffset(long loeschOffset)
  {
    _loeschOffset = loeschOffset;
  }

  public static int getGroesseInfoAnteil()
  {
    return _groesseInfoAnteil;
  }

  public void setGroesseInfoAnteil(int groesseInfoAnteil)
  {
    _groesseInfoAnteil = groesseInfoAnteil;
  }

  public static ConfigurationObject getVersandObjekt()
  {
    return _versandObjektKex;
  }

  public static void setVersandObjekt(ConfigurationObject versandObjekt)
  {
    _versandObjektKex = versandObjekt;
  }

  public static Vector<ResultData> getInputDaten()
  {
    return _inputDaten;
  }

  public static void setInputDaten(Vector<ResultData> inputDaten)
  {
    _inputDaten = inputDaten;
  }

  public static Map<String, Set<String>> getMeldungsGruppenZuordnung()
  {
    return _meldungsGruppenZuordnung;
  }

  public static void setMeldungsGruppenZuordnung(Map<String, Set<String>> meldungsGruppenZuordnung)
  {
    _meldungsGruppenZuordnung = meldungsGruppenZuordnung;
  }

  public static long getNextLaufendeNummerEmailFax()
  {
    return _laufendeNummerEmailFax += 1;
  }
}
