/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.bmvew.bmvew.regel;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.Data.TextArray;
import de.bsvrz.dav.daf.main.Data.TextValue;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.vew.bmvew.bmvew.BetriebsMeldungApp;
import de.bsvrz.vew.bmvew.bmvew.plugin.StandardFilter;

/**
 * Klasse implementiert die Filterfunktionalitaet
 * 
 * Anpassungen NERZ-nderungsantrag 34 (Meldungstext wird ausgewertet)
 * 
 * @version $Revision: 1.2 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
 * @author Dambach-Werke GmbH
 * @author Timo Pittner, Uwe Peuker
 * 
 */
public class Filter
{
  /**
   * Der Text des Kriterium
   */
  TextValue kriterium;

  /**
   * Der Text der Operation
   */
  TextValue operation;

  /**
   * Der Text des Werts
   */
  TextArray wert;

  /**
   * Der Text der Plug-In-Operation
   */
  TextValue operationPlugin;

  /**
   * Der Text der Plug-In-Wertes
   */
  TextArray wertPlugin;

  /**
   * Das Package in welchem sich die Plugin-Klassen befinden
   */
  String pluginPackage;

  /**
   * Die definierten Kriterien
   * 
   * @version $Revision: 1.2 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
   * 
   * @author Dambach-Werke GmbH
   * @author Timo Pittner
   * 
   */
  public static enum Kriterien {
    Meldungstyp("Meldungstyp"), Meldungsklasse("Meldungklasse"), Applikation("Applikation"), Referenz("Referenz"), Benutzer(
        "Benutzer"), Veranlasser("Veranlasser"), Ursache("Ursache"), erweitert("erweitert"), Meldungstext(
        "Meldungstext");

    private final String wert;

    private Kriterien(final String s)
    {
      wert = s;
    }

    private String getWert()
    {

      return wert;
    }

  }

  /**
   * Die definierten Operationen
   * 
   * @version $Revision: 1.2 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
   * 
   * @author Dambach-Werke GmbH
   * @author Timo Pittner
   * 
   */
  public static enum Operationen {
    Wertistgleich("Wert ist gleich"), Wertistkleiner("Wert ist kleiner"), Wertistgrer("Wert ist grer"), Wertistkleinerodergleich(
        "Wert ist kleiner oder gleich"), Wertistgrerodergleich("Wert ist grer oder gleich"), Textgleich(
        "Text gleich"), Textenthlt("Text enthlt"), Textbeginntmit("Text beginnt mit"), Textendetmit("Text endet mit"), keinOperator(
        "kein Operator");

    private final String wert;

    private Operationen(final String s)
    {
      wert = s;
    }

    private String getWert()
    {

      return wert;
    }
  }

  /**
   * Die definierten Meldungstypen
   * 
   * @version $Revision: 1.2 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
   * 
   * @author Dambach-Werke GmbH
   * @author Timo Pittner
   * 
   */
  public static enum MeldungsTyp {
    Fach("Fach"), System("System");

    private final String wert;

    private MeldungsTyp(final String s)
    {
      wert = s;
    }

  }

  /**
   * Die definierten Meldungsklassen
   * 
   * @version $Revision: 1.2 $ / $Date: 2015/06/09 13:02:12 $ / ($Author: Pittner $)
   * 
   * @author Dambach-Werke GmbH
   * @author Timo Pittner
   * 
   */
  public static enum MeldungsKlasse {
    Fatal("Fatal"), Fehler("Fehler"), Warnung("Warnung"), Information("Information");

    private final String wert;

    private MeldungsKlasse(final String s)
    {
      wert = s;
    }
  }

  /**
   * Initialisiert den Filter
   * 
   * @param kriterium
   *          der Text des Kriteriums
   * @param op
   *          der Text der Operation
   * @param wert
   *          der Text des Wertes
   * @throws IllegalArgumentException
   */
  public void initialisiere(final TextValue kriterium, final TextValue op, final TextArray wert)
      throws IllegalArgumentException
  {

    if ((kriterium == null) || (op == null) || (wert == null))
    {

      throw new IllegalArgumentException();
    }

    switch (Kriterien.valueOf(kriterium.getText()))
    {
      case Meldungstyp:

        for (int i = 0; i < wert.getLength(); i++)
        {

          switch (MeldungsTyp.valueOf(wert.getText(i)))
          {

            case Fach:
            case System:

              break;

            default:
              throw new IllegalArgumentException();

          }

        }

        break;

      case Meldungsklasse:

        for (int i = 0; i < wert.getLength(); i++)
        {

          switch (MeldungsKlasse.valueOf(wert.getText(i)))
          {

            case Fatal:
            case Fehler:
            case Warnung:
            case Information:

              break;

            default:
              throw new IllegalArgumentException();

          }

        }

        break;

      case Applikation:
      case Referenz:
      case Benutzer:
      case Veranlasser:
      case Ursache:
      case erweitert:
      case Meldungstext:
        break;

      default:
        throw new IllegalArgumentException();

    }

    // Das Enum verarbeitet den bergeben Text ohne Leerzeichen, h!?
    final String ohneLeerzeichen = op.getText().replace(" ", "");

    switch (Operationen.valueOf(ohneLeerzeichen))
    {

      case Wertistgleich:
      case Wertistkleiner:
      case Wertistgrer:
      case Wertistkleinerodergleich:
      case Wertistgrerodergleich:
      case Textgleich:
      case Textenthlt:
      case Textbeginntmit:
      case Textendetmit:
      case keinOperator:
        break;

      default:

        throw new IllegalArgumentException();
    }

    for (int i = 0; i < wert.getLength(); i++)
    {
      final String text = wert.getText(i);

      if (text.length() == 0)
      {
        throw new IllegalArgumentException();
      }

    }

    this.kriterium = kriterium;
    this.operation = op;
    this.wert = wert;

  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString()
  {
    // TODO Auto-generated method stub
    final StringBuffer buffer = new StringBuffer();
    buffer.append("Wert:");
    for (int i = 0; i < wert.getLength(); i++)
    {
      buffer.append("[" + wert.getText(i) + "]");
    }

    final String s = "Kriterium:" + kriterium.getText() + "; Operation:" + operation.getText() + "; " + buffer;
    return s;
  }

  /**
   * Fhrt den Filter aus
   * 
   * @param betriebsMeldung
   * 
   * @return true, wenn die Betriebsmeldung dem Filter entspricht
   */
  public boolean pruefe(final Data betriebsMeldung)
  {

    final Boolean[] aktion = new Boolean[wert.getLength()];

    String value = null;

    switch (Kriterien.valueOf(kriterium.getText()))
    {
      case Meldungstyp:
        value = betriebsMeldung.getItem("MeldungsTyp").valueToString();
        break;

      case Meldungsklasse:
        value = betriebsMeldung.getItem("MeldungsKlasse").valueToString();
        break;

      case Meldungstext:
        value = betriebsMeldung.getItem("MeldungsText").valueToString();
        break;

      case Applikation:
        value = betriebsMeldung.getItem("ApplikationsKennung").valueToString();
        break;

      case Referenz:
        value = betriebsMeldung.getItem("Referenz").valueToString();
        break;

      case Benutzer:
        value = betriebsMeldung.getItem("Urlasser").getItem("BenutzerReferenz").valueToString();
        break;

      case Veranlasser:
        value = betriebsMeldung.getItem("Urlasser").getItem("Veranlasser").valueToString();
        break;

      case Ursache:
        value = betriebsMeldung.getItem("Urlasser").getItem("Ursache").valueToString();
        break;

      // Hier werden die Plugins abgearbeitet
      case erweitert:

        final AttributeGroup atg = BetriebsMeldungApp.getConnection().getDataModel()
            .getAttributeGroup("atg.betriebsMeldungsVerwaltungRegel");
        final Data data = BetriebsMeldungApp.getConnection().createData(atg);

        final Array a1 = data.getArray("Regel");
        a1.setLength(1);
        final Array a2 = a1.getItem(0).getArray("Filter");
        a2.setLength(1);
        // Data d1 = a2.getItem(0).getItem("Kriterium");
        final Data d1 = a2.getItem(0);

        wertPlugin = d1.getItem("Wert").asTextArray();
        wertPlugin.setLength(wert.getLength() - 1);

        operationPlugin = d1.getItem("Operation").asTextValue();
        operationPlugin.setText(operation.getText());

        for (int i = 0; i < wert.getLength(); i++)
        {
          final String text = wert.getText(i);

          if (i == 0)
          {

            pluginPackage = text;

          }
          else
          {

            wertPlugin.getTextValue(i - 1).setText(text);

          }

        }

        try
        {
          final Class<?> klass = Class.forName(pluginPackage);
          try
          {
            final StandardFilter filterPlugin = (StandardFilter)klass.newInstance();

            filterPlugin.initialisiere(operationPlugin, wertPlugin);

            if (filterPlugin.pruefe(betriebsMeldung))
            {
              return true;
            }
            else
            {
              return false;
            }
          }
          catch (final InstantiationException e)
          {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
          catch (final IllegalAccessException e)
          {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }

        }
        catch (final ClassNotFoundException e)
        {
          e.printStackTrace();
        }

        break;

      default:

    }

    switch (Operationen.valueOf(operation.getText().replace(" ", "")))
    {

      case Wertistgleich:

        try
        {
          for (int i = 0; i < wert.getLength(); i++)
          {
            final String text = wert.getText(i);
            final Long l1 = new Long(value);
            final Long l2 = new Long(text);
            if (l1 == l2)
            {
              aktion[i] = true;
            }
            else
            {
              aktion[i] = false;
            }
          }

        }
        catch (final NumberFormatException e)
        {
          // TODO Auto-generated catch block
          aktion[0] = false;
        }

        break;

      case Wertistkleiner:

        try
        {

          for (int i = 0; i < wert.getLength(); i++)
          {
            final String text = wert.getText(i);
            final Long l1 = new Long(value);
            final Long l2 = new Long(text);
            if (l1 < l2)
            {
              aktion[i] = true;
            }
            else
            {
              aktion[i] = false;
            }
          }
        }
        catch (final NumberFormatException e)
        {
          // TODO Auto-generated catch block
          aktion[0] = false;
        }

        break;

      case Wertistgrer:

        try
        {

          for (int i = 0; i < wert.getLength(); i++)
          {
            final String text = wert.getText(i);
            final Long l1 = new Long(value);
            final Long l2 = new Long(text);
            if (l1 > l2)
            {
              aktion[i] = true;
            }
            else
            {
              aktion[i] = false;
            }
          }
        }
        catch (final NumberFormatException e)
        {
          // TODO Auto-generated catch block
          aktion[0] = false;
        }

        break;

      case Wertistkleinerodergleich:

        try
        {
          for (int i = 0; i < wert.getLength(); i++)
          {
            final String text = wert.getText(i);
            final Long l1 = new Long(value);
            final Long l2 = new Long(text);
            if (l1 <= l2)
            {
              aktion[i] = true;
            }
            else
            {
              aktion[i] = false;
            }
          }
        }
        catch (final NumberFormatException e)
        {
          // TODO Auto-generated catch block
          aktion[0] = false;
        }

        break;

      case Wertistgrerodergleich:

        try
        {

          for (int i = 0; i < wert.getLength(); i++)
          {
            final String text = wert.getText(i);
            final Long l1 = new Long(value);
            final Long l2 = new Long(text);
            if (l1 >= l2)
            {
              aktion[i] = true;
            }
            else
            {
              aktion[i] = false;
            }
          }

        }
        catch (final NumberFormatException e)
        {
          // TODO Auto-generated catch block
          aktion[0] = false;
        }

        break;

      case Textgleich:

        for (int i = 0; i < wert.getLength(); i++)
        {
          final String text = wert.getText(i);
          if (value.equals(text))
          {
            aktion[i] = true;
          }
          else
          {
            aktion[i] = false;
          }
        }

        break;

      case Textenthlt:

        for (int i = 0; i < wert.getLength(); i++)
        {
          final String text = wert.getText(i);
          if (value.contains(text))
          {
            aktion[i] = true;
          }
          else
          {
            aktion[i] = false;
          }
        }

        break;

      case Textbeginntmit:

        for (int i = 0; i < wert.getLength(); i++)
        {
          final String text = wert.getText(i);
          if (value.startsWith(text))
          {
            aktion[i] = true;
          }
          else
          {
            aktion[i] = false;
          }
        }

        break;

      case Textendetmit:

        for (int i = 0; i < wert.getLength(); i++)
        {
          final String text = wert.getText(i);
          if (value.endsWith(text))
          {
            aktion[i] = true;
          }
          else
          {
            aktion[i] = false;
          }
        }

        break;

      case keinOperator:

        for (int i = 0; i < wert.getLength(); i++)
        {
          aktion[i] = false;
        }

        break;

      default:

    }

    // Die Werte werden Oder-Verknuepft!
    for (final Boolean element : aktion)
    {
      if (element)
      {
        return true;
      }
    }

    return false;
  }

}
